import torch
from datasets import load_dataset
from torch.utils.data import DataLoader


arg_map = {
    "glue_mnli": ("glue", "mnli"),
    "glue_sst-2": ("glue", "sst2"),
    "glue_mrpc": ("glue", "mrpc"),
    "glue_cola": ("glue", "cola"),
    "glue_qnli": ("glue", "qnli"),
    "glue_qqp": ("glue", "qqp"),
    "glue_rte": ("glue", "rte"),
    "glue_sts-b": ("glue", "stsb"),
    "glue_wnli": ("glue", "wnli")
}


def get_dataloader(args, tokenizer):
    train_dataset, eval_dataset = load_data(args, tokenizer)

    shuffle = True if args.shuffle == "true" else False
    train_dataloader = DataLoader(train_dataset, batch_size=args.batch_size, shuffle=shuffle)
    eval_dataloader = DataLoader(eval_dataset, batch_size=args.batch_size)

    return train_dataloader, eval_dataloader


def load_data(args, tokenizer):
    if args.mode == "pretraining":
        return load_data_pretrain(args, tokenizer)
    elif args.mode == "finetuning":
        return load_data_finetune(args, tokenizer)
    else:
        raise ValueError("Invalid mode. Choose 'pretraining' or 'finetuning'")


def load_data_pretrain(args, tokenizer):
    def prepare_dataset(dataset):
        def tokenize_function_pretrain(batch):
            encoding = tokenizer(batch["text"], truncation=True, padding="max_length", max_length=args.max_length)
            return {
                "input_ids": torch.tensor(encoding["input_ids"]).clone().detach().to(torch.long),
                "attention_mask": torch.tensor(encoding["attention_mask"]).clone().detach().to(torch.long),
            }

        dataset = dataset.map(tokenize_function_pretrain, batched=True, remove_columns=["text", "timestamp", "url"])
        dataset.with_format("torch")

        return dataset

    train_dataset = load_dataset("c4", "en", split="train", streaming=True)
    # TODO Does it make a difference in memory to define this after or before training?
    eval_dataset = load_dataset("c4", "en", split="validation", streaming=True)

    train_dataset = prepare_dataset(train_dataset)
    eval_dataset = prepare_dataset(eval_dataset)

    return train_dataset, eval_dataset


def load_data_finetune(args, tokenizer):
    task_to_keys = {
        "glue_cola": ("sentence", None),
        "glue_mnli": ("premise", "hypothesis"),
        "glue_mrpc": ("sentence1", "sentence2"),
        "glue_qnli": ("question", "sentence"),
        "glue_qqp": ("question1", "question2"),
        "glue_rte": ("sentence1", "sentence2"),
        "glue_sst-2": ("sentence", None),
        "glue_sts-b": ("sentence1", "sentence2"),
        "glue_wnli": ("sentence1", "sentence2"),
    }
    if args.dataset not in arg_map:
        raise ValueError(f"Data set '{args.dataset}' not supported for mode 'finetuning'!")
    dataset = load_dataset(*arg_map[args.dataset])

    # Extract useful text
    sentence1_key, sentence2_key = task_to_keys[args.dataset]

    def tokenize_function_finetune(batch):
        texts = (
            (batch[sentence1_key],) if sentence2_key is None else (batch[sentence1_key], batch[sentence2_key])
        )
        return tokenizer(*texts, truncation=True, padding="max_length", max_length=args.max_length)
    
    dataset = dataset.map(tokenize_function_finetune)
    dataset.set_format(type="torch", columns=["input_ids", "attention_mask", "label"])

    eval_dataset = dataset["validation_matched" if args.task_name == "glue_mnli" else "validation"]
    train_dataset = dataset["train"]

    return train_dataset, eval_dataset
