from torch.optim.lr_scheduler import LinearLR

from load_data import get_dataloader, arg_map
from load_models import get_model
from load_optimizers import get_optimizer
from load_lr_scheduler import get_scheduler

from math import ceil
from logger import init_csv, log_to_csv
from accelerate import Accelerator
import torch
import evaluate
from args import args
from time import perf_counter


def train(accelerator, scheduler, model, optimizer, train_dataloader, eval_dataloader, num_epochs):
    """ training model """
    model, optimizer, train_dataloader = accelerator.prepare(model, optimizer, train_dataloader)  # To GPU
    model.train()
    
    for epoch in range(num_epochs):
        total_loss = 0
        batch_cnt = 0
        for batch in train_dataloader:
            if args.num_training_steps is not None and args.num_training_steps <= batch_cnt:
                break

            start_time = perf_counter()
            optimizer.zero_grad()
            
            outputs = model(**batch)

            loss = outputs.loss
            accelerator.backward(loss)
            optimizer.step()
            scheduler.step()

            batch_cnt += 1
            
            compute_time = perf_counter() - start_time

            if batch_cnt % args.eval_every == 0:
                train_loss = loss.item()
                eval_loss, eval_mertic = evaluate(accelerator, model, tokenizer, eval_dataloader)
                log_to_csv(epoch + 1, batch_cnt, compute_time, train_loss, eval_loss, eval_mertic)

            total_loss += loss.item()
        
        # avg_loss = total_loss / max(1, batch_cnt)
        # print(f"Epoch {epoch+1}, Avg. training loss: {avg_loss:.4f}")
    
    return model


def evaluate(accelerator, model, tokenizer, eval_dataloader):
    if args.mode == "finetuning":
        metric = evaluate.load(*arg_map[args.task_name])
    else:
        metric = None  # TODO Metric fro pretraining

    model.eval()

    total_loss = torch.tensor(0.0).to(device)
    perplexity = 0  # Only for pretraining
    num_batches = len(train_dataloader) if args.num_training_steps is None else args.num_training_steps

    # eval_dataloader.batch = lambda batch_size: training_utils.batch_fn(eval_dataloader, batch_size)
    # for batch in eval_dataloader.batch(batch_size=args.batch_size):
    for step, batch in enumerate(eval_dataloader):
        if args.num_eval_steps is not None and args.num_eval_steps <= step:
            break

        if args.mode == "finetuning":
            is_regression = args.task_name == "glue_stsb"
            labels = batch["labels"]

            with torch.no_grad():
                outputs = model(**batch, labels=labels)  # model(**batch)

            predictions = outputs.logits.argmax(dim=-1) if not is_regression else outputs.logits.squeeze()
            predictions, references = accelerator.gather((predictions, labels))

            metric.add_batch(
                predictions=predictions,
                references=references,
            )
        else:
            batch = {k: v.to(device) for k, v in batch.items()}
            labels = batch["input_ids"].clone()
            labels[labels == tokenizer.pad_token_id] = -100

            with torch.no_grad():
                outputs = model(**batch, labels=labels)

            perplexity += torch.exp(outputs.loss).item()

        loss = outputs.loss
        total_loss += loss.detach()

    total_loss = total_loss / num_batches
    if args.mode == "finetuning":
        eval_metric = metric.compute()
    else:
        eval_metric = perplexity / num_batches

    return total_loss, eval_metric


if __name__ == "__main__":
    if args.test == "true":
        print("Test mode")
        accelerator = Accelerator()
    else:
        accelerator = Accelerator(mixed_precision="bf16")

    device = "cuda" if torch.cuda.is_available() else "cpu"

    print(f"Running on: {device}")
    print(f"Using optimizer: {args.optimizer}")
    print(f"Arguments: {args}")
    init_csv()

    # Load stuff
    model, tokenizer = get_model(args)

    train_dataloader, eval_dataloader = get_dataloader(args, tokenizer)

    # Init setup
    optimizer, model = get_optimizer(args, model)

    num_batches = len(train_dataloader) if args.num_training_steps is None else args.num_training_steps
    num_steps = args.num_epochs * num_batches
    lr_scheduler = get_scheduler(
        optimizer, args.lr_scheduler, args.warm_up_fraction, num_steps, args.lr, args.lr_min
    )

    # Prepare everything with our `accelerator`.
    model, optimizer, train_dataloader, eval_dataloader, lr_scheduler = accelerator.prepare(
        model, optimizer, train_dataloader, eval_dataloader, lr_scheduler
    )

    # Train
    trained_model = train(accelerator, lr_scheduler, model, optimizer, train_dataloader, num_epochs=args.num_epochs)

    # Evaluate

    file_name = f"{args.model}_{args.optimizer}_pretrained" if args.mode == "pretraining" else f"{args.model}_{args.optimizer}_finetuned"
    model_path = f"models/{file_name}"
    trained_model.save_pretrained(model_path)
    tokenizer.save_pretrained(model_path)